package params

// MonthsPerYear is the number of months in a year.
const MonthsPerYear = 12

// WeeksPerMonth is the number of weeks in a month.
const WeeksPerMonth = 4

// LutID is an identifier for land use types.
type LutID uint8

// Enumeration for land use types using iota
const (
	Forest LutID = iota
	Agriculture
	Urban
	lutIDEnd
)

// SpeciesID is an identifier for a species.
type SpeciesID uint8

// Enumeration for species using iota
const (
	Limpet SpeciesID = iota
	Mayfly
	Chironomid
	speciesEnd
)

// MortalityModes defines the mortality modes for the model.
type MortalityModes struct {
	GeneralMortality     bool // Indicates if general mortality is active
	TemperatureMortality bool // Indicates if temperature-related mortality is active
	DriftMortality       bool // Indicates if drift mortality is active
	AgingMortality       bool // Indicates if aging mortality is active
	ZeroOutMortality     bool // Indicates if zero-out mortality is active
}

// BiofilmParameters contains parameters to calculate the biofilm growth.
type BiofilmParameters struct {
	GrowthRate               float64 // Growth rate of biofilm
	MortalityRate            float64 // Mortality rate of biofilm
	CoreCapacity             float64 // Capacity of base and thickSusceptible, just used at the beginning
	CapacityBase             float64 // Capacity of base
	CapacityThickSusceptible float64 // Capacity of thickSusceptible
	CapacityMicroNiche       float64 // Capacity of micro niche
	NonEdibleBiofilmMass     float64 // Amount of biofilm which is not edible
	FractionBase             float64 // Fraction of capacity that is for base biofilm
	FractionThickSusceptible float64 // Fraction of capacity that is for thickSusceptible biofilm
	FractionMicroNiche       float64 // Fraction of overall biofilm mass that is in micro niche
	Beta0                    float64 // Lower threshold for eutrophication
	Beta1                    float64 // Upper threshold for eutrophication
	MaxPropabilityFila       float64 // Maximum probability of filamentous algae growth [-]
	ThresholdFila            float64 // Threshold for filamentous algae growth [g DW/segment]
	LifeSpanFila             int64   // Life span of filamentous algae [weeks]
}

// SolarRadiationParameters contains parameters to calculate light response.
type SolarRadiationParameters struct {
	MeanLogDistr       float64 // Mean for cumulative log normal distribution (light response curve)
	StdLogDistr        float64 // Standard deviation for cumulative log normal distribution (light response curve)
	PAR                float64 // Fraction of light that is photosynthetic active radiation
	Scaling            float64 // Scaling factor from kWh/m² per month (solar radiation) to μmol/m²s
	ShadingFactor      float64 // Shading factor for shaded riversides (forests)
	MaximumTreeFoliage float64 // Maximum foliage of trees
}

// ModelParams contains the parameters for the model.
type ModelParams struct {
	MortalitySwitches MortalityModes           // Switches to turn on/off the five mortality modes
	FilaSwitch        bool                     // Switch to turn on/off the filamentous algae growth
	Biofilm           BiofilmParameters        // Parameters to calculate biofilm growth
	SolarRadiation    SolarRadiationParameters // Parameters to calculate light response
}

// newModelParams creates a new ModelParams with default values.
func newModelParams() ModelParams {
	return ModelParams{
		MortalitySwitches: MortalityModes{
			GeneralMortality:     true,
			TemperatureMortality: true,
			DriftMortality:       true,
			AgingMortality:       true,
			ZeroOutMortality:     true,
		},
		FilaSwitch: true,
		Biofilm: BiofilmParameters{
			GrowthRate:               0.71,
			MortalityRate:            0.01,
			CoreCapacity:             402000,
			CapacityBase:             0,
			CapacityThickSusceptible: 0,
			CapacityMicroNiche:       30000,
			NonEdibleBiofilmMass:     1800,
			FractionBase:             0.8,
			FractionThickSusceptible: 0.1,
			FractionMicroNiche:       0.1,
			Beta0:                    300000,
			Beta1:                    350000,
			MaxPropabilityFila:       0.5,
			ThresholdFila:            25000,
			LifeSpanFila:             12,
		},
		SolarRadiation: SolarRadiationParameters{
			MeanLogDistr:       5,
			StdLogDistr:        1,
			PAR:                0.5,
			Scaling:            6.3889,
			ShadingFactor:      0.6,
			MaximumTreeFoliage: 5.46333,
		},
	}
}
