package params

// DispersalParameters contains parameters to calculate species dispersal.
type DispersalParameters struct {
	ProbStay     float64
	ProbUp       float64
	ProbDown     float64
	DistMeanUp   float64
	DistMeanDown float64
}

// TempMortalityParameters contains parameters for temperature mortality.
type TempMortalityParameters struct {
	Threshold float64
	Slope     float64
	Intercept float64
}

// SpeciesParams contains parameters for a species.
type SpeciesParams struct {
	ID   SpeciesID // Identifier for the species
	Name string    // Name of the species

	// Reproduction
	OvipositionMonth uint8               // Month when the species oviposit
	Offspring        uint32              // Amount of offspring per mature individual
	Dispersal        DispersalParameters // Dispersal parameters: probStay, probDown, probUp, DistMeanUp, DistMeanDown. See dispersalID

	// Development
	HatchingMonth uint8 // Month when the eggs hatch (become larvae)
	MatureMonth   uint8 // Month when the larvae become mature. AdultM in the Java Version

	// Mortality
	MortalityRate       float64                 // Mortality rate for general mortality
	WinterScalingFactor float64                 // Winter scaling for general mortality
	TempMortality       TempMortalityParameters // Threshold, slope, intercept for temperature mortality
	DriftMortalityRate  float64                 // Mortality rate for drifting species
	DeathMonth          uint8                   // Month when the species dies due to aging mortality
}

// NewSpeciesParameters creates a new SpeciesParams instance based on the provided SpeciesID with default values.
func newSpeciesParameters(spID SpeciesID) SpeciesParams {
	switch spID {
	case Limpet:
		return SpeciesParams{
			ID:                  Limpet,
			Name:                "Limpet",
			OvipositionMonth:    4,
			HatchingMonth:       6,
			MatureMonth:         4,
			DeathMonth:          6,
			Offspring:           50,
			MortalityRate:       0.04,
			WinterScalingFactor: 1,
			DriftMortalityRate:  1.0,
			TempMortality: TempMortalityParameters{
				Threshold: 18.0,
				Slope:     -0.055,
				Intercept: 1.09,
			},
		}
	case Mayfly:
		return SpeciesParams{
			ID:                  Mayfly,
			Name:                "Mayfly",
			OvipositionMonth:    4,
			HatchingMonth:       7,
			MatureMonth:         4,
			DeathMonth:          5,
			Offspring:           200,
			MortalityRate:       0.07,
			WinterScalingFactor: 1,
			Dispersal: DispersalParameters{
				ProbStay:     0.04,
				ProbDown:     0.9,
				ProbUp:       0.1,
				DistMeanUp:   17.0,
				DistMeanDown: 2.0,
			},
			TempMortality: TempMortalityParameters{
				Threshold: 16.0,
				Slope:     -0.055,
				Intercept: 1.88,
			},
			DriftMortalityRate: 0.05,
		}
	case Chironomid:
		return SpeciesParams{
			ID:                  Chironomid,
			Name:                "Choronomid",
			OvipositionMonth:    10,
			HatchingMonth:       1,
			MatureMonth:         4,
			DeathMonth:          5,
			Offspring:           25,
			MortalityRate:       0.28,
			WinterScalingFactor: 1,
			Dispersal: DispersalParameters{
				ProbStay:     0.15,
				ProbDown:     0.6,
				ProbUp:       0.4,
				DistMeanUp:   3.0,
				DistMeanDown: 2.0,
			},
			TempMortality: TempMortalityParameters{
				Threshold: 13.0,
				Slope:     -0.005,
				Intercept: 1.065,
			},
			DriftMortalityRate: 0.05,
		}
	default:
		panic("No valid SpeciesID")
	}
}
