package sys

import (
	"math"

	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"gonum.org/v1/gonum/stat/distuv"

	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark/ecs"
)

// BiofilmGrowth holds the resources for the BiofilmGrowth system.
type BiofilmGrowth struct {
	paramSetRes ecs.Resource[params.ParameterSet]
	modelRes    ecs.Resource[model.Model]
	tickRes     ecs.Resource[resource.Tick]
}

// Initialize the resources for the BiofilmGrowth system and set up the biofilm
// capacities and initial biofilm mass by including fractions.
func (b *BiofilmGrowth) Initialize(w *ecs.World) {
	b.paramSetRes = ecs.NewResource[params.ParameterSet](w)
	b.modelRes = ecs.NewResource[model.Model](w)
	b.tickRes = ecs.NewResource[resource.Tick](w)

	paramSet := b.paramSetRes.Get()
	mod := b.modelRes.Get()

	biofilmParams := &paramSet.Mod.Biofilm

	if biofilmParams.FractionBase+biofilmParams.FractionThickSusceptible+biofilmParams.FractionMicroNiche != 1 {
		panic("FractionBase, FractionThickSusceptible and FractionMicroNiche must sum to 1")
	}

	// Setting the capacities of the core biofilm compartments (base and thick
	// suscecptible)
	biofilmParams.CapacityBase = biofilmParams.CoreCapacity * (biofilmParams.FractionBase / (biofilmParams.FractionBase + biofilmParams.FractionThickSusceptible))
	biofilmParams.CapacityThickSusceptible = biofilmParams.CoreCapacity * (biofilmParams.FractionThickSusceptible / (biofilmParams.FractionBase + biofilmParams.FractionThickSusceptible))

	// Setting the initial biofilm mass of the compartments
	for i := range mod.Segments {
		seg := &mod.Segments[i]

		seg.BaseBiofilm = biofilmParams.NonEdibleBiofilmMass * biofilmParams.FractionBase
		seg.ThickSusceptibleBiofilm = biofilmParams.NonEdibleBiofilmMass * biofilmParams.FractionThickSusceptible
		seg.MicroNicheBiofilm = biofilmParams.NonEdibleBiofilmMass * biofilmParams.FractionMicroNiche
	}
}

// Update the BiofilmGrowth system. This function calculates the effective
// reproduction rate, effective capacity of all compartments, and updates the
// biofilm mass of all compartments in each segment.
func (b *BiofilmGrowth) Update(w *ecs.World) {
	paramSet := b.paramSetRes.Get()
	mod := b.modelRes.Get()
	tick := b.tickRes.Get().Tick

	biofilmParams := &paramSet.Mod.Biofilm
	solarRadParams := &paramSet.Mod.SolarRadiation

	for i := range mod.Segments {
		seg := &mod.Segments[i]

		effReproduction := effectiveReproduction(biofilmParams, solarRadParams, paramSet.Seg[i].LandUseType, tick)
		effCapacityBase := effectiveCapacity(biofilmParams, solarRadParams, biofilmParams.CapacityBase, paramSet.Seg[i].LandUseType, tick)
		effCapacityThickPossible := effectiveCapacity(biofilmParams, solarRadParams, biofilmParams.CapacityThickSusceptible, paramSet.Seg[i].LandUseType, tick)
		effCapacityMicroNiche := effectiveCapacity(biofilmParams, solarRadParams, biofilmParams.CapacityMicroNiche, paramSet.Seg[i].LandUseType, tick)

		seg.BaseBiofilm = calculateBiofilm(seg.BaseBiofilm, effReproduction, effCapacityBase)
		seg.ThickSusceptibleBiofilm = calculateBiofilm(seg.ThickSusceptibleBiofilm, effReproduction, effCapacityThickPossible)
		seg.MicroNicheBiofilm = calculateBiofilm(seg.MicroNicheBiofilm, effReproduction, effCapacityMicroNiche)
	}
}

// Finalize BiofilmGrowth.
func (b *BiofilmGrowth) Finalize(w *ecs.World) {}

// effectiveReproduction calculates the effective reproduction rate of biofilm
// by subtracting the mortality rate from the effective growth rate.
func effectiveReproduction(biofilmParams *params.BiofilmParameters, solarRadParams *params.SolarRadiationParameters, lut params.LutID, tick int64) float64 {
	return effectiveGrowthRate(biofilmParams, solarRadParams, lut, tick) - biofilmParams.MortalityRate
}

// effectiveGrowthRate calculates the effective growth rate of biofilm by
// multiplying the growth rate with the light effect and nutrient effect.
func effectiveGrowthRate(biofilmParams *params.BiofilmParameters, solarRadParams *params.SolarRadiationParameters, lut params.LutID, tick int64) float64 {
	return biofilmParams.GrowthRate * lightEffect(solarRadParams, lut, tick) * nutrientEffect()
}

// effectiveCapacity calculates the effective capacity of a biofilm compartment
// by considering the effective growth rate, mortality rate and the capacity of
// the compartment.
func effectiveCapacity(biofilmParams *params.BiofilmParameters, solarRadParams *params.SolarRadiationParameters, capacity float64, lut params.LutID, tick int64) float64 {
	effGrowthRate := effectiveGrowthRate(biofilmParams, solarRadParams, lut, tick)

	if effGrowthRate == 0 {
		return 0
	}

	return (1 - (biofilmParams.MortalityRate / effGrowthRate)) * capacity
}

// lightEffect calculates the light effect on biofilm growth using a log-normal
// distribution. It considers the solar radiation, scaling factor, PAR, and
// seasonal shading.
func lightEffect(solarRadParams *params.SolarRadiationParameters, lut params.LutID, tick int64) float64 {
	radiation := solarRadiation(tick)
	pfd := radiation * solarRadParams.Scaling * solarRadParams.PAR * (1 - seasonalShade(lut, solarRadParams, tick))
	logNormal := distuv.LogNormal{Mu: solarRadParams.MeanLogDistr, Sigma: solarRadParams.StdLogDistr}

	return logNormal.CDF(pfd)
}

// solarRadiation calculates the solar radiation based on the tick value. It
// uses a cosine function to simulate the daily variation of solar radiation.
func solarRadiation(tick int64) float64 {
	return 87.9 - 78.8*math.Cos(2*math.Pi*((float64(tick)+1.4)/48))
}

// seasonalShade calculates the seasonal shading effect on solar radiation based
// on the land use type and the tick value. It uses a quadratic function to
// simulate the shading effect of trees in a forested area.
func seasonalShade(lut params.LutID, solarRadParams *params.SolarRadiationParameters, tick int64) float64 {
	if lut != params.Forest {
		return 0.0
	}

	treeFoliage := -7.87 + 0.12*float64(tick) - 0.00027*math.Pow(float64(tick), 2)

	if treeFoliage < 0 {
		return 0.0
	}

	return solarRadParams.ShadingFactor * treeFoliage / solarRadParams.MaximumTreeFoliage
}

// nutrientEffect calculates the nutrient effect on biofilm growth. It is a
// placeholder function that currently returns a constant value of 1.0.
func nutrientEffect() float64 {
	return 1.0
}

// calculateBiofilm calculates the biofilm mass in a compartment based on the
// current biofilm mass, reproduction rate, and effective capacity of the
// compartment.
func calculateBiofilm(current, reproduction, capacity float64) float64 {
	if current == 0 {
		return 0
	}

	return 1 / (math.Exp(-reproduction)/current + (1-math.Exp(-reproduction))/capacity)
}
