package sys

import (
	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark/ecs"
)

// Development holds the resources for the Development system.
type Development struct {
	paramSetRes ecs.Resource[params.ParameterSet]
	modelRes    ecs.Resource[model.Model]
	tickRes     ecs.Resource[resource.Tick]
}

// Initialize the resources for the Development system.
func (d *Development) Initialize(w *ecs.World) {
	d.paramSetRes = ecs.NewResource[params.ParameterSet](w)
	d.modelRes = ecs.NewResource[model.Model](w)
	d.tickRes = ecs.NewResource[resource.Tick](w)
}

// Update the Development system. Iterate over segments and species and execute
// hatching and becomeMature functions.
func (d *Development) Update(w *ecs.World) {
	tick := d.tickRes.Get().Tick

	// Leave this system if it is not the first week of the month
	if tick%4 != 0 {
		return
	}

	paramSet := d.paramSetRes.Get()
	mod := d.modelRes.Get()

	for i := range mod.Segments {
		seg := &mod.Segments[i]

		for id := range seg.SpeciesSet {
			sp := &seg.SpeciesSet[id]

			hatching(sp, paramSet.Spec[id].HatchingMonth, tick)
			becomeMature(sp, paramSet.Spec[id].MatureMonth, tick)
		}
	}
}

// Finalize Development.
func (d *Development) Finalize(w *ecs.World) {}

// hatching checks if the current month is the hatching month for the species
// and if so, it adds the abundance of eggs to the abundance of larvae and sets
// the abundance of eggs to 0.
func hatching(sp *model.Species, hatchingMonth uint8, tick int64) {
	if uint8((tick/params.WeeksPerMonth)%params.MonthsPerYear) == hatchingMonth {
		sp.Abundance[model.Larvae] += sp.Abundance[model.Egg]
		sp.Abundance[model.Egg] = 0
	}
}

// becomeMature checks if the current month is the mature month for the species
// and if so, it adds the abundance of larvae to the abundance of mature
// individuals and sets the abundance of larvae to 0.
func becomeMature(sp *model.Species, matureMonth uint8, tick int64) {
	if uint8((tick/params.WeeksPerMonth)%params.MonthsPerYear) == matureMonth {
		sp.Abundance[model.Mature] += sp.Abundance[model.Larvae]
		sp.Abundance[model.Larvae] = 0
	}
}
