package sys

import (
	"math/rand/v2"

	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark/ecs"
)

// Eutrophication holds the resources for the Eutrophication system.
type Eutrophication struct {
	paramSetRes ecs.Resource[params.ParameterSet]
	modelRes    ecs.Resource[model.Model]
	tickRes     ecs.Resource[resource.Tick]
}

// Initialize the resources for the Eutrophication system.
func (e *Eutrophication) Initialize(w *ecs.World) {
	e.paramSetRes = ecs.NewResource[params.ParameterSet](w)
	e.modelRes = ecs.NewResource[model.Model](w)
	e.tickRes = ecs.NewResource[resource.Tick](w)
}

// Update the Eutrophication system. For every segment, check if biofilm rips
// off and if thick biofilm or filamentous algae grow.
func (e *Eutrophication) Update(w *ecs.World) {
	paramSet := e.paramSetRes.Get()
	mod := e.modelRes.Get()
	tick := e.tickRes.Get().Tick

	for i := range mod.Segments {
		seg := &mod.Segments[i]

		checkBiofilmRipOff(seg, paramSet.Mod.Biofilm)
		checkThickBiofilm(seg, paramSet.Mod.Biofilm)
		checkFila(seg, paramSet.Mod.Biofilm, paramSet.Mod.FilaSwitch, tick)
	}
}

// Finalize Eutrophication.
func (e *Eutrophication) Finalize(w *ecs.World) {}

// checkBiofilmRipOff checks if the biofilm mass of the base and thick
// susceptible compartement exceed the threshold beta0 and if so, it rips off
// the excess biofilm mass. The left over biofilm mass is calculated with the
// non edible biofilm parameter.
func checkBiofilmRipOff(seg *model.Segment, par params.BiofilmParameters) {
	// Base biofilm
	if seg.BaseBiofilm > par.Beta1*par.FractionBase {
		seg.RippedBaseBiofilm = seg.BaseBiofilm - par.NonEdibleBiofilmMass*par.FractionBase
		seg.BaseBiofilm = par.NonEdibleBiofilmMass * par.FractionBase
	}

	// Thick susceptible biofilm
	if seg.ThickSusceptibleBiofilm > par.Beta1*par.FractionThickSusceptible {
		seg.RippedThickSusceptibleBiofilm = seg.ThickSusceptibleBiofilm - par.NonEdibleBiofilmMass*par.FractionThickSusceptible
		seg.ThickSusceptibleBiofilm = par.NonEdibleBiofilmMass * par.FractionThickSusceptible
	}
}

// checkThickBiofilm checks if the biofilm mass of the base and thick
// susceptible compartments exceed the threshold beta0 and if so, it sets the
// eutrophic status of the segment to true.
func checkThickBiofilm(seg *model.Segment, par params.BiofilmParameters) {
	// Base biofilm
	seg.IsBaseEutrophic = seg.BaseBiofilm > par.Beta0*par.FractionBase

	// Thick susceptible biofilm
	seg.IsThickSusceptibleEutrophic = seg.ThickSusceptibleBiofilm > par.Beta0*par.FractionThickSusceptible
}

// checkFila checks if filamentous algae grow.
func checkFila(seg *model.Segment, par params.BiofilmParameters, filaSwitch bool, tick int64) {
	// Check first if the filamentous algea just died, if so, set the base
	// biofilm to the non edible mass and return
	if tick == seg.DeathTickFila {
		seg.BaseBiofilm = par.NonEdibleBiofilmMass * par.FractionBase
		return
	}

	// If switch for filamentous algea growth is off, it is not the first week of the month,
	// the base biofilm is below the threshold, the segment is already eutrophic or filamentous
	// algea already exist, return
	if !filaSwitch || tick%4 != 0 || seg.BaseBiofilm < par.ThresholdFila || seg.IsBaseEutrophic || tick < seg.DeathTickFila {
		return
	}

	// Check if filamentous algae grow.
	if rand.Float64() < calculateFilaPropability(seg.BaseBiofilm, par.MaxPropabilityFila, par.ThresholdFila, par.Beta0) {
		// If filamentous algae grow, set the death tick to the current tick
		// plus the lifespan of filamentous algae
		seg.DeathTickFila = tick + par.LifeSpanFila
		// The base biofilm is set to zero during the growth of filamentous
		// algae
		seg.BaseBiofilm = 0
	}
}

// calculateFilaPropability returns a value between 0 and 1 which represents the
// propability if filamentous algae growth. The value is calculated by a linear
// increase function dependend on the biofilm mass of the base compartment.
func calculateFilaPropability(baseBio, maxProb, thresholdFila, beta0 float64) float64 {
	return (maxProb/(beta0-thresholdFila))*baseBio - (maxProb/(beta0-thresholdFila))*thresholdFila
}
