package sys

import (
	"testing"

	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"github.com/mlange-42/ark-tools/app"
	"github.com/mlange-42/ark-tools/system"
	"github.com/mlange-42/ark/ecs"
	"github.com/stretchr/testify/assert"
)

func TestEutrophication(t *testing.T) {
	app := app.New()
	m := model.New(1)
	paramSet := params.NewParameterSet(1)

	// Setup state variables
	m.Segments[0].BaseBiofilm = 10.0
	m.Segments[0].ThickSusceptibleBiofilm = 10.0

	// Setup parameters
	paramSet.Mod.Biofilm.FractionBase = 0.8
	paramSet.Mod.Biofilm.FractionThickSusceptible = 0.2
	paramSet.Mod.Biofilm.FractionMicroNiche = 0.0
	paramSet.Mod.Biofilm.Beta0 = 10.0 // Base biofilm is higher than this threshold: 10 > 10*0.9
	paramSet.Mod.Biofilm.Beta1 = 20.0 // Thick susceptible biofilm is higher than this threshold: 10 > 20*0.1
	paramSet.Mod.Biofilm.NonEdibleBiofilmMass = 0.0

	ecs.AddResource(&app.World, &paramSet)
	ecs.AddResource(&app.World, &m)

	// Add systems
	app.AddSystem(&Eutrophication{})
	app.AddSystem(&system.FixedTermination{Steps: 1})

	// Run the simulation
	app.Run()

	// Base compartment is eutrophic
	assert.True(t, m.Segments[0].IsBaseEutrophic)
	// Thick suceptible compartment was etrophic then ripped off and is not etrophic anymore
	assert.False(t, m.Segments[0].IsThickSusceptibleEutrophic)
	// Because non edible biofilm mass is 0, the ripped off biofilm mass is equal to the initial base biofilm mass
	assert.Equal(t, 10.0, m.Segments[0].RippedThickSusceptibleBiofilm)
}

func TestCheckFila(t *testing.T) {
	mod := model.New(1)
	paramSet := params.NewParameterSet(1)

	// Test case: Filamentous algae grow
	paramSet.Mod.Biofilm.Beta0 = 10.0
	paramSet.Mod.Biofilm.MaxPropabilityFila = 1.0
	paramSet.Mod.Biofilm.ThresholdFila = 1.0
	paramSet.Mod.Biofilm.LifeSpanFila = 10
	tick := int64(4) // First week of the month

	mod.Segments[0].BaseBiofilm = 10.0 // Setting to beta0 to get highest probability

	checkFila(&(mod.Segments[0]), paramSet.Mod.Biofilm, true, tick)
	assert.Equal(t, 0.0, mod.Segments[0].BaseBiofilm)         // Base biofilm is set to 0 during the growth of filamentous algae
	assert.Equal(t, int64(14), mod.Segments[0].DeathTickFila) // Death tick is set to current tick + life span of filamentous algae

	// Test case: Filamentous algae just died
	expectedBaseBiofilm := paramSet.Mod.Biofilm.NonEdibleBiofilmMass * paramSet.Mod.Biofilm.FractionBase

	mod.Segments[0].DeathTickFila = 12
	tick = 12

	checkFila(&(mod.Segments[0]), paramSet.Mod.Biofilm, true, tick)
	assert.Equal(t, expectedBaseBiofilm, mod.Segments[0].BaseBiofilm) // Base biofilm is set to non edible mass
	assert.Equal(t, int64(12), mod.Segments[0].DeathTickFila)         // Death tick is not changed

	// Test case: Filamentous algae do not grow
	mod.Segments[0].BaseBiofilm = 20.0
	mod.Segments[0].DeathTickFila = 6
	tick = 4

	checkFila(&(mod.Segments[0]), paramSet.Mod.Biofilm, true, tick)
	assert.Equal(t, 20.0, mod.Segments[0].BaseBiofilm) // Base biofilm is not changed
}
