package sys

import (
	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark/ecs"
)

// Mortality holds the resources for the Mortality system.
type Mortality struct {
	paramSetRes ecs.Resource[params.ParameterSet]
	modelRes    ecs.Resource[model.Model]
	tickRes     ecs.Resource[resource.Tick]
}

// Initialize the resources for the Mortality system.
func (m *Mortality) Initialize(w *ecs.World) {
	m.paramSetRes = ecs.NewResource[params.ParameterSet](w)
	m.modelRes = ecs.NewResource[model.Model](w)
	m.tickRes = ecs.NewResource[resource.Tick](w)
}

// Update the Mortality system. This function iterates over all segments and
// species and applies general mortality, temperature mortality, drift
// mortality, aging mortality and zero out small abundance mortality.
func (m *Mortality) Update(w *ecs.World) {
	paramSet := m.paramSetRes.Get()
	mod := m.modelRes.Get()
	tick := m.tickRes.Get().Tick

	for i := range mod.Segments {
		seg := &mod.Segments[i]

		for id := range seg.SpeciesSet {
			spec := &seg.SpeciesSet[id]
			specParams := paramSet.Spec[id]
			month := (tick / params.WeeksPerMonth) % params.MonthsPerYear

			// General, temperature and drift mortality applies for larvae of
			// all species (water-bound)...
			generalMortality(spec, model.Larvae, specParams.MortalityRate, specParams.WinterScalingFactor, tick, paramSet.Mod.MortalitySwitches.GeneralMortality)
			temperatureMortality(spec, model.Larvae, paramSet.Seg[i].TempWater[month], specParams.TempMortality, paramSet.Mod.MortalitySwitches.TemperatureMortality)
			driftMortality(spec, model.Larvae, specParams.DriftMortalityRate, paramSet.Mod.MortalitySwitches.DriftMortality)

			// ...but just for mature limpets (also water-bound)
			if id == int(params.Limpet) {
				generalMortality(spec, model.Mature, specParams.MortalityRate, specParams.WinterScalingFactor, tick, paramSet.Mod.MortalitySwitches.GeneralMortality)
				temperatureMortality(spec, model.Mature, paramSet.Seg[i].TempWater[month], specParams.TempMortality, paramSet.Mod.MortalitySwitches.TemperatureMortality)
				driftMortality(spec, model.Mature, specParams.DriftMortalityRate, paramSet.Mod.MortalitySwitches.DriftMortality)
			}

			agingMortality(spec, specParams.DeathMonth, tick, paramSet.Mod.MortalitySwitches.AgingMortality)
			zeroOutSmallAbundance(spec, paramSet.Mod.MortalitySwitches.ZeroOutMortality)
		}
	}
}

// Finalize Mortality.
func (m *Mortality) Finalize(w *ecs.World) {}

// generalMortality applies a general mortality rate to the abundance of a
// species at a given life stage. It also applies a winter scaling factor if the
// current month is December, January or February. Panics if the abundance
// becomes negative.
func generalMortality(species *model.Species, ls model.LifeStage, mortalityRate, winterScalingFactor float64, tick int64, mSwitch bool) {
	if !mSwitch {
		return
	}

	scalingFactor := 1.0

	if month := (tick / params.WeeksPerMonth) % params.MonthsPerYear; month == 11 || month == 0 || month == 1 {
		scalingFactor = winterScalingFactor
	}

	species.Abundance[ls] *= (1 - scalingFactor*mortalityRate)

	if species.Abundance[ls] < 0 {
		panic("Negative Abundance in generalMortality")
	}
}

// temperatureMortality applies a temperature mortality function to the
// abundance of a species at a given life stage. Panics if the abundance becomes
// negative.
func temperatureMortality(species *model.Species, ls model.LifeStage, temp float64, params params.TempMortalityParameters, mSwitch bool) {
	if !mSwitch {
		return
	}

	species.Abundance[ls] *= temperatureMortalityFunction(temp, params.Threshold, params.Slope, params.Intercept)

	if species.Abundance[ls] < 0 {
		panic("Negative Abundance in temperatureMortality")
	}
}

// temperatureMortalityFunction returns a value between 0-1 when temperature is
// lower than the threshold. The return value is calculated by a simple linear
// function.
func temperatureMortalityFunction(temp, threshold, slope, intercept float64) float64 {
	if temp < threshold {
		return 1.0
	}

	return slope*temp + intercept
}

// driftMortality applies drift mortality rate to immigrants. Panics if the
// abundance becomes negative.
func driftMortality(species *model.Species, ls model.LifeStage, driftMortalityRate float64, mSwitch bool) {
	if !mSwitch {
		return
	}

	species.Immigrants[ls] *= (1 - driftMortalityRate)

	if species.Abundance[ls] < 0 {
		panic("Negative Abundance in driftMortality")
	}
}

// agingMortality sets abundance of mature individuals to zero when death month
// is reached.
func agingMortality(species *model.Species, deathMonth uint8, tick int64, mSwitch bool) {
	if !mSwitch {
		return
	}

	if (tick/params.WeeksPerMonth)%params.MonthsPerYear == int64(deathMonth) {
		species.Abundance[model.Mature] = 0.0
	}
}

// zeroOutSmallAbundance sets abundances of all life stages to zero if they
// become too small.
func zeroOutSmallAbundance(species *model.Species, mSwitch bool) {
	if !mSwitch {
		return
	}

	thresKill := 0.00001

	for i := range species.Abundance {
		if species.Abundance[i] < thresKill {
			species.Abundance[i] = 0.0
		}
	}
}
