package sys

import (
	"testing"

	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"github.com/mlange-42/ark-tools/app"
	"github.com/mlange-42/ark-tools/system"
	"github.com/mlange-42/ark/ecs"
	"github.com/stretchr/testify/assert"
)

func TestGeneralMortality(t *testing.T) {
	app := app.New()
	paramSet := params.NewParameterSet(model.DefaultSegNumber)
	m := model.New(model.DefaultSegNumber)

	// Setting parameters
	mortalityRate := 0.5
	var steps int64 = 5

	paramSet.Mod.MortalitySwitches = params.MortalityModes{
		GeneralMortality:     true,
		TemperatureMortality: false,
		DriftMortality:       false,
		AgingMortality:       false,
		ZeroOutMortality:     false,
	}
	paramSet.Spec[params.Limpet].MortalityRate = mortalityRate
	paramSet.Spec[params.Mayfly].MortalityRate = mortalityRate
	paramSet.Spec[params.Chironomid].MortalityRate = mortalityRate

	ecs.AddResource(&app.World, &m)
	ecs.AddResource(&app.World, &paramSet)
	app.AddSystem(&Mortality{})

	app.AddSystem(&system.FixedTermination{Steps: steps})
	app.Run()

	// limpets initial abundance: 20000; 20000*(0.5**5)=625; Eggs are not affected
	assert.Equal(t, []float64{20000, 625, 625}, m.Segments[0].SpeciesSet[params.Limpet].Abundance)
	// mayflies initial abundance: 10000; 10000*(0.5**5)=312.5; Eggs and mature individuals are not affected
	assert.Equal(t, []float64{10000, 312.5, 10000}, m.Segments[0].SpeciesSet[params.Mayfly].Abundance)
	// chironomids initial abundance: 300000; 300000*(0.5**5)=3125; Eggs and mature individuals are not affected
	assert.Equal(t, []float64{100000, 3125, 100000}, m.Segments[0].SpeciesSet[params.Chironomid].Abundance)
}

func TestTemperatureMortality(t *testing.T) {
	app := app.New()
	paramSet := params.NewParameterSet(model.DefaultSegNumber)
	m := model.New(model.DefaultSegNumber)

	// Setting parameters
	tempMort := params.TempMortalityParameters{
		Threshold: 10,
		Slope:     -0.05,
		Intercept: 1,
	}
	var steps int64 = 5

	paramSet.Mod.MortalitySwitches = params.MortalityModes{
		GeneralMortality:     false,
		TemperatureMortality: true,
		DriftMortality:       false,
		AgingMortality:       false,
		ZeroOutMortality:     false,
	}
	paramSet.Spec[params.Limpet].TempMortality = tempMort
	paramSet.Spec[params.Mayfly].TempMortality = tempMort
	paramSet.Spec[params.Chironomid].TempMortality = tempMort

	ecs.AddResource(&app.World, &m)
	ecs.AddResource(&app.World, &paramSet)
	app.AddSystem(&Mortality{})

	app.AddSystem(&system.FixedTermination{Steps: steps})
	app.Run()

	// limpets initial abundance: 20000; 20000*(0.5**5)=625; Eggs are not affected
	assert.Equal(t, []float64{20000, 625, 625}, m.Segments[0].SpeciesSet[params.Limpet].Abundance)
	// mayflies initial abundance: 10000; 10000*(0.5**5)=312.5; Eggs and mature individuals are not affected
	assert.Equal(t, []float64{10000, 312.5, 10000}, m.Segments[0].SpeciesSet[params.Mayfly].Abundance)
	// chironomids initial abundance: 300000; 300000*(0.5**5)=3125; Eggs and mature individuals are not affected
	assert.Equal(t, []float64{100000, 3125, 100000}, m.Segments[0].SpeciesSet[params.Chironomid].Abundance)
}

func TestDriftMortality(t *testing.T) {
	app := app.New()
	paramSet := params.NewParameterSet(model.DefaultSegNumber)
	m := model.New(model.DefaultSegNumber)

	// Setting parameters
	mortalityRate := 0.5
	var steps int64 = 5

	paramSet.Mod.MortalitySwitches = params.MortalityModes{
		GeneralMortality:     false,
		TemperatureMortality: false,
		DriftMortality:       true,
		AgingMortality:       false,
		ZeroOutMortality:     false,
	}
	paramSet.Spec[params.Limpet].DriftMortalityRate = 1 // They don't drift and just die
	paramSet.Spec[params.Mayfly].DriftMortalityRate = mortalityRate
	paramSet.Spec[params.Chironomid].DriftMortalityRate = mortalityRate

	ecs.AddResource(&app.World, &m)
	ecs.AddResource(&app.World, &paramSet)
	app.AddSystem(&Mortality{})

	app.AddSystem(&system.FixedTermination{Steps: steps})
	app.Run()

	// limpets initial abundance: 2000; 2000*(0.5**5)=625; Eggs are not affected
	assert.Equal(t, []float64{2000, 0, 0}, m.Segments[0].SpeciesSet[params.Limpet].Immigrants)
	// mayflies initial abundance: 1000; 1000*(0.5**5)=31.25; Eggs and mature individuals are not affected
	assert.Equal(t, []float64{1000, 31.25, 1000}, m.Segments[0].SpeciesSet[params.Mayfly].Immigrants)
	// chironomids initial abundance: 3000; 30000*(0.5**5)=312.5; Eggs and mature individuals are not affected
	assert.Equal(t, []float64{10000, 312.5, 10000}, m.Segments[0].SpeciesSet[params.Chironomid].Immigrants)
}

func TestAgingMortality(t *testing.T) {
	app := app.New()
	paramSet := params.NewParameterSet(model.DefaultSegNumber)
	m := model.New(model.DefaultSegNumber)

	// Setting parameters
	paramSet.Mod.MortalitySwitches = params.MortalityModes{
		GeneralMortality:     false,
		TemperatureMortality: false,
		DriftMortality:       false,
		AgingMortality:       true,
		ZeroOutMortality:     false,
	}

	ecs.AddResource(&app.World, &m)
	ecs.AddResource(&app.World, &paramSet)
	app.AddSystem(&Mortality{})

	app.AddSystem(&system.FixedTermination{Steps: 48})
	app.Run()

	// Mature individuals die at DeathMonth, so after a year no mature individual should be alive
	assert.Equal(t, []float64{20000, 20000, 0}, m.Segments[0].SpeciesSet[params.Limpet].Abundance)
	assert.Equal(t, []float64{10000, 10000, 0}, m.Segments[0].SpeciesSet[params.Mayfly].Abundance)
	assert.Equal(t, []float64{100000, 100000, 0}, m.Segments[0].SpeciesSet[params.Chironomid].Abundance)
}

func TestZeroOutSmallMortality(t *testing.T) {
	app := app.New()
	paramSet := params.NewParameterSet(model.DefaultSegNumber)
	m := model.New(model.DefaultSegNumber)

	// Setting Parameters
	mortalityRate := 0.1
	var steps int64 = 1000

	paramSet.Mod.MortalitySwitches = params.MortalityModes{
		GeneralMortality:     true,
		TemperatureMortality: false,
		DriftMortality:       false,
		AgingMortality:       false,
		ZeroOutMortality:     true,
	}
	paramSet.Spec[params.Limpet].MortalityRate = mortalityRate
	paramSet.Spec[params.Mayfly].MortalityRate = mortalityRate
	paramSet.Spec[params.Chironomid].MortalityRate = mortalityRate

	ecs.AddResource(&app.World, &m)
	ecs.AddResource(&app.World, &paramSet)
	app.AddSystem(&Mortality{})

	app.AddSystem(&system.FixedTermination{Steps: steps})
	app.Run()

	// After 1000 steps, abundance should be zero; Eggs are not affected
	assert.Equal(t, []float64{20000, 0, 0}, m.Segments[0].SpeciesSet[params.Limpet].Abundance)
	assert.Equal(t, []float64{10000, 0, 10000}, m.Segments[0].SpeciesSet[params.Mayfly].Abundance)
	assert.Equal(t, []float64{100000, 0, 100000}, m.Segments[0].SpeciesSet[params.Chironomid].Abundance)
}
