package sys

import (
	"math"

	"git.ufz.de/oesa/go-mastiff/model"
	"git.ufz.de/oesa/go-mastiff/params"
	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark/ecs"
)

// Reproduction holds the resources for the Reproduction system.
type Reproduction struct {
	paramSetRes ecs.Resource[params.ParameterSet]
	modelRes    ecs.Resource[model.Model]
	tickRes     ecs.Resource[resource.Tick]

	Dispersed []float64 // Slice for dispersed imagines
}

// Initialize the resources and a slice to store the dispersed individuals for
// the Reproduction system.
func (r *Reproduction) Initialize(w *ecs.World) {
	r.paramSetRes = ecs.NewResource[params.ParameterSet](w)
	r.modelRes = ecs.NewResource[model.Model](w)
	r.tickRes = ecs.NewResource[resource.Tick](w)

	paramSet := r.paramSetRes.Get()
	r.Dispersed = make([]float64, paramSet.NumberOfSegments)
}

// Update the Reproduction system. Iterate over segments and species and execute
// oviposition function for limpets and dispersal and oviposition functions for
// mayflies and chironomids.
func (r *Reproduction) Update(w *ecs.World) {
	tick := r.tickRes.Get().Tick

	// Leave this system if it is not the first week of the month
	if tick%4 != 0 {
		return
	}

	mod := r.modelRes.Get()
	paramSet := r.paramSetRes.Get()

	// Limpets do simple reproduction
	simpleReproduction(mod.Segments, params.Limpet, paramSet.Spec[params.Limpet].Offspring, paramSet.Spec[params.Limpet].OvipositionMonth, tick)

	// Mayflies and chironomids disperse and then reproduce
	dispersal(mod.Segments, r.Dispersed, params.Mayfly, paramSet.Spec[params.Mayfly], tick)
	dispersal(mod.Segments, r.Dispersed, params.Chironomid, paramSet.Spec[params.Chironomid], tick)
}

// Finalize Reproduction.
func (r *Reproduction) Finalize(w *ecs.World) {}

// simpleReproduction checks if the current month is the oviposition month for
// the species and if so, it executes the oviposition function for all segments.
func simpleReproduction(segments []model.Segment, spID params.SpeciesID, offspring uint32, ovipositionMonth uint8, tick int64) {
	if uint8((tick/params.WeeksPerMonth)%params.MonthsPerYear) != ovipositionMonth {
		return
	}

	for i := range segments {
		oviposition(&segments[i].SpeciesSet[spID], offspring)
	}
}

// dispersal checks if the current month is the oviposition month for the
// species and if so, it first collects the abundance of mature individuals in
// each segment and then distributes them according to the dispersal
// probability. Finally, it executes the oviposition function for all segments.
func dispersal(segments []model.Segment, dispersed []float64, spID params.SpeciesID, spParams params.SpeciesParams, tick int64) {
	if uint8((tick/params.WeeksPerMonth)%params.MonthsPerYear) != spParams.OvipositionMonth {
		return
	}

	// Collect the abundance of mature individuals in each segment
	for i := range segments {
		seg := &segments[i]

		dispersed[i] = seg.SpeciesSet[spID].Abundance[model.Mature]
		seg.SpeciesSet[spID].Abundance[model.Mature] = 0
	}

	// Disperse
	for taker := range segments {
		seg := &segments[taker]

		for donor := range segments {
			segIndexDiff := donor - taker
			seg.SpeciesSet[spID].Abundance[model.Mature] += dispersed[donor] * dispersalProbability(segIndexDiff, spParams.Dispersal.ProbStay, spParams.Dispersal.ProbUp, spParams.Dispersal.ProbDown, spParams.Dispersal.DistMeanUp, spParams.Dispersal.DistMeanDown)
		}

		// Place eggs
		oviposition(&seg.SpeciesSet[spID], spParams.Offspring)
	}
}

// dispersalProbability calculates the probability of dispersal based on the
// distance between segments and the dispersal parameters. It uses a decay
// function to determine the probability of dispersal.
func dispersalProbability(segIndexDiff int, probStay, probUp, probDown, distMeanUp, distMeanDown float64) float64 {
	if segIndexDiff < 0 {
		return (1 - probStay) * probDown * dispersalDecayFunction(segIndexDiff, distMeanDown)
	} else if segIndexDiff > 0 {
		return (1 - probStay) * probUp * dispersalDecayFunction(segIndexDiff, distMeanUp)
	} else {
		return probStay
	}
}

// dispersalDecayFunction returns the decay function for the dispersal
// probability. It uses the exponential decay function to calculate the
// probability of dispersal based on the distance between segments and the
// dispersal parameters.
func dispersalDecayFunction(segIndexDiff int, distMean float64) float64 {
	return (1 / distMean) * math.Exp(-(1/distMean)*math.Abs(float64(segIndexDiff)))
}

// oviposition calculates the number of eggs laid by the species. It adds the
// abundance of mature individuals multiplied by the number of offspring to the
// abundance of eggs.
func oviposition(species *model.Species, offspring uint32) {
	species.Abundance[model.Egg] += species.Abundance[model.Mature] * float64(offspring)
}
